import time
from PyQt5.QtCore import QVariant
from qgis.core import QgsApplication,QgsCoordinateReferenceSystem,QgsCoordinateTransform,QgsFeature,QgsField,QgsFields,QgsGeometry,QgsMapLayer,QgsPointXY,QgsProject,QgsTask,QgsVectorLayer,QgsWkbTypes
from qgis.gui import QgisInterface,QgsMapCanvas
from ..com.orbitgis.tool.message.Message import Message
from ..orbitgt.mapping3d.sdk.viewer.AList import AList
from ..orbitgt.mapping3d.sdk.viewer.Bounds import Bounds
from ..orbitgt.mapping3d.sdk.viewer.MapObject import MapObject
from ..orbitgt.mapping3d.sdk.viewer.MeasureState import MeasureState
from ..orbitgt.mapping3d.sdk.viewer.Viewer3DM import Viewer3DM
from .MapObjectLayer import MapObjectLayer
from .MapObjectRenderer import MapObjectRenderer
from .MapObjects import MapObjects

# Class UpdateMeasurement manages the updates to the measurement layer
#
class UpdateMeasurement:

    # Create a new updater
    #
    def __init__(self,viewer: Viewer3DM,layer: MapObjectLayer,iface: QgisInterface,plugin):
        self.viewer = viewer
        self.layer = layer
        self.iface = iface
        self.plugin = plugin
        self.updateTime = 0.0
        self.lastRequestTime = 0.0
        self.isUpdating = False
        self.measureState = None
        self.mapObjects = None

    # Called when the measurement has been changed and the display need to be updated
    #
    def updatedMeasurement(self):
        self.updateTime = time.time()

    # Check for updates
    #
    def checkForUpdates(self) -> None:
        # already updating?
        if (self.isUpdating):
            return
        # already up-to-date?
        if (self.lastRequestTime>self.updateTime):
            return
        currentTime = time.time()
        # throttle the rate of update requests
        if (self.lastRequestTime>0.0) and (currentTime<self.lastRequestTime+0.1):
            return
        # throttle the rate of measure changes
        if (self.updateTime>0.0) and (currentTime<self.updateTime+0.1):
            return
        # start the update task (to free the calling timer thread)
        self.isUpdating = True
        self.lastRequestTime = currentTime
        self.updateTask = QgsTask.fromFunction("Update measurement",self.update1,on_finished=self.update2)
        QgsApplication.taskManager().addTask(self.updateTask)

    # Update (in background thread)
    #
    def update1(self,task: QgsTask):
        try:
            Message.print(__name__,"Requesting measurement")
            self.measureState = self.viewer.getMeasureState()
            if (self.measureState!=None): self.mapObjects = self.measureState.mapObjects
            if (self.mapObjects!=None): Message.print(__name__,"Got "+str(self.mapObjects.size())+" mapobjects")
            self.isUpdating = False
            return
        except Exception as e:
            Message.printError(__name__,e,"Error while updating measurement")
            self.isUpdating = False
            return
    
    # Finish the update (called by the main UI thread)
    #
    def update2(self,taskException):
        self.layer.update(self.mapObjects,True)

    # Set the value of a named attribute of the new feature object
    #
    def setAttribute(self,feature: QgsFeature,attributeFields: QgsFields,propertyName: str,propertyValue: str):
        if (propertyName==None):
            return
        if (propertyValue==None):
            return
        attributeIndex = attributeFields.indexOf(propertyName)
        if (attributeIndex<0):
            return
        attributeField = attributeFields.at(attributeIndex) # QgsField
        attributeValue = QVariant(propertyValue)
        Message.print(__name__,"Found attribute '"+propertyName+"'")
        converted = attributeValue.convert(attributeField.type())
        if (converted):
            feature.setAttribute(attributeIndex,attributeValue)
            Message.print(__name__,"Initialized attribute '"+propertyName+"' with '"+propertyValue+"'")
    
    # Save (add) the current measurement to the active layer
    #
    def doSaveMeasurement(self) -> None:
        Message.print(__name__,"Adding the measurement to the current layer")
        # we need to have a measurement
        if (self.measureState==None):
            Message.print(__name__,"There is no current measurement")
            return
        # print the attributes
        if (self.measureState.properties!=None):
            Message.print(__name__,"The properties of the measurement are:")
            for i in range(0,self.measureState.properties.size()):
                property = self.measureState.properties.get(i)
                if (property.name!=None): Message.print(__name__,">name: '"+property.name+"'")
                if (property.label!=None): Message.print(__name__,"label: '"+property.label+"'")
                if (property.value!=None): Message.print(__name__,"value: '"+property.value+"'")
                if (property.unit!=None): Message.print(__name__,"unit: '"+property.unit+"'")
        # get the current vector layer
        currentLayer = self.iface.mapCanvas().currentLayer()
        if (currentLayer==None):
            Message.print(__name__,"There is no current layer to add to")
            return
        if (isinstance(currentLayer,QgsVectorLayer)==False):
            Message.print(__name__,"The current layer '"+currentLayer.name()+"' is not a vector layer")
            return
        if (currentLayer.isEditable()==False):
            Message.print(__name__,"The current layer '"+currentLayer.name()+"' has not been made editable")
            return
        # create the geometry for the layer
        layerGeometryType = QgsWkbTypes.geometryType(currentLayer.wkbType())
        geometry = None
        # get the (first) mapobject of the measurement
        if (self.measureState.mapObjects.size()<1):
            Message.print(__name__,"The measure state does not have a mapobject")
            return
        mapObject = self.measureState.mapObjects.get(0)
        if (layerGeometryType==QgsWkbTypes.PointGeometry):
            if (mapObject.spatialPoint==None):
                Message.print(__name__,"Cannot add a non-point measurement to a point layer")
                return
            Message.print(__name__,"Creating a point geometry")
            geometry = MapObjects.getPointGeometry(mapObject)
        if (layerGeometryType==QgsWkbTypes.LineGeometry):
            if (mapObject.spatialLine==None):
                Message.print(__name__,"Cannot add a non-line measurement to a line layer")
                return
            Message.print(__name__,"Creating a line geometry")
            geometry = MapObjects.getLineGeometry(mapObject)
        if (layerGeometryType==QgsWkbTypes.PolygonGeometry):
            if (mapObject.spatialArea==None):
                Message.print(__name__,"Cannot add a non-area measurement to an area layer")
                return
            Message.print(__name__,"Creating an area geometry")
            geometry = MapObjects.getAreaGeometry(mapObject)
        if (geometry==None):
            Message.print(__name__,"The current layer geometry is neither point, line or polygon")
            return
        # transform the geometry from the mapobject CRS to the layer CRS
        Message.print(__name__,"Transforming the geometry to the current layer crs")
        mapObjectToLayer = QgsCoordinateTransform(QgsCoordinateReferenceSystem("EPSG:"+MapObjects.CRS),currentLayer.sourceCrs(),QgsProject.instance())
        geometry.transform(mapObjectToLayer)
        # make new features for the layer
        Message.print(__name__,"Creating new features")
        features = []
        feature = QgsFeature()
        feature.setGeometry(geometry)
        feature.setFields(currentLayer.fields(),True)
        attributeFields = feature.fields() # QgsFields
        # set any default attribute values from the application configuration
        configuration = self.plugin.readConfigurationFile()
        configurationKeys = configuration.listKeys()
        for i in range(0,configurationKeys.size()):
            key = configurationKeys.get(i)
            if (key.startswith("default.attribute.value.")):
                attributeName = key[24:].strip(" ")
                attributeValue = configuration.get(key)
                Message.print(__name__,"Applying default attribute '"+attributeName+"' value '"+attributeValue+"'")
                self.setAttribute(feature,attributeFields,attributeName,attributeValue);
        # set attribute values from the measure state
        if (self.measureState.properties!=None):
            for i in range(0,self.measureState.properties.size()):
                property = self.measureState.properties.get(i)
                Message.print(__name__,"Applying mesure property '" + property.name + "' value '" + property.value + "'")
                self.setAttribute(feature,attributeFields,property.name,property.value)
        features.append(feature)
        # add the features to the layer
        Message.print(__name__,"Adding the new features to editable layer '"+currentLayer.name()+"'")
        #currentLayer.startEditing()
        for feature in features: currentLayer.addFeature(feature)
        #committed = currentLayer.commitChanges()
        #if (committed): Message.print(__name__,"The features were committed")
        #else: Message.print(__name__,"The features were not committed")
        currentLayer.updateExtents()
        self.iface.mapCanvas().refresh()
